from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.db.models import BinaryField, DateField
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.datetime_safe import date

from djangoldp.fields import LDPUrlField
from djangoldp.models import Model
from djangoldp.permissions import LDPPermissions
from djangoldp.tests.permissions import HalfRandomPermissions


class User(AbstractUser, Model):

    class Meta(AbstractUser.Meta, Model.Meta):
        serializer_fields = ['@id', 'username', 'first_name', 'last_name', 'email', 'userprofile',
                             'conversation_set', 'circle_set', 'projects']
        anonymous_perms = ['view', 'add']
        authenticated_perms = ['inherit', 'change']
        owner_perms = ['inherit']


class Skill(Model):
    title = models.CharField(max_length=255, blank=True, null=True)
    obligatoire = models.CharField(max_length=255)
    slug = models.SlugField(blank=True, null=True, unique=True)
    date = models.DateTimeField(auto_now_add=True, blank=True)

    def recent_jobs(self):
        return self.joboffer_set.filter(date__gte=date.today())

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']
        serializer_fields = ["@id", "title", "recent_jobs", "slug"]
        lookup_field = 'slug'


class JobOffer(Model):
    title = models.CharField(max_length=255, null=True)
    skills = models.ManyToManyField(Skill, blank=True)
    slug = models.SlugField(blank=True, null=True, unique=True)
    date = models.DateTimeField(auto_now_add=True, blank=True)

    def recent_skills(self):
        return self.skills.filter(date__gte=date.today())

    def some_skill(self):
        return self.skills.all().first()

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'change', 'add']
        owner_perms = ['inherit', 'delete', 'control']
        serializer_fields = ["@id", "title", "skills", "recent_skills", "resources", "slug", "some_skill", "urlid"]
        container_path = "job-offers/"
        lookup_field = 'slug'


class Conversation(models.Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    author_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING)
    peer_user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name="peers_conv",
                                  on_delete=models.DO_NOTHING)

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']
        owner_field = 'author_user'


class Resource(Model):
    joboffers = models.ManyToManyField(JobOffer, blank=True, related_name='resources')
    description = models.CharField(max_length=255)

    class Meta(Model.Meta):
        anonymous_perms = ['view', 'add', 'delete', 'change', 'control']
        authenticated_perms = ['inherit']
        owner_perms = ['inherit']
        serializer_fields = ["@id", "joboffers"]
        depth = 1


class UserProfile(Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    user = models.OneToOneField(settings.AUTH_USER_MODEL, related_name='userprofile', on_delete=models.CASCADE)

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit']
        owner_perms = ['inherit', 'change', 'control']
        depth = 1


class Message(models.Model):
    text = models.CharField(max_length=255, blank=True, null=True)
    conversation = models.ForeignKey(Conversation, on_delete=models.DO_NOTHING)
    author_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING)

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']


class Dummy(models.Model):
    some = models.CharField(max_length=255, blank=True, null=True)
    slug = models.SlugField(blank=True, null=True, unique=True)

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']


class LDPDummy(Model):
    some = models.CharField(max_length=255, blank=True, null=True)

    class Meta(Model.Meta):
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']


# model used in django-guardian permission tests (no anonymous etc permissions set)
class PermissionlessDummy(Model):
    some = models.CharField(max_length=255, blank=True, null=True)
    slug = models.SlugField(blank=True, null=True, unique=True)
    parent = models.ForeignKey(LDPDummy, on_delete=models.DO_NOTHING, related_name="anons", blank=True, null=True)

    class Meta(Model.Meta):
        anonymous_perms = []
        authenticated_perms = []
        owner_perms = []
        permissions = (
            ('custom_permission_permissionlessdummy', 'Custom Permission'),
        )


class Post(Model):
    content = models.CharField(max_length=255)
    author = models.ForeignKey(UserProfile, blank=True, null=True, on_delete=models.SET_NULL)
    peer_user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name="peers_post",
                                  on_delete=models.SET_NULL)

    class Meta(Model.Meta):
        auto_author = 'author'
        auto_author_field = 'userprofile'
        anonymous_perms = ['view', 'add', 'delete', 'add', 'change', 'control']
        authenticated_perms = ['inherit']
        owner_perms = ['inherit']


class Invoice(Model):
    title = models.CharField(max_length=255, blank=True, null=True)
    date = models.DateField(blank=True, null=True)

    class Meta(Model.Meta):
        depth = 2
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']


class Circle(Model):
    name = models.CharField(max_length=255, blank=True)
    description = models.CharField(max_length=255, blank=True)
    team = models.ManyToManyField(settings.AUTH_USER_MODEL, through="CircleMember", blank=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="owned_circles", on_delete=models.DO_NOTHING, null=True, blank=True)

    class Meta(Model.Meta):
        anonymous_perms = ['view', 'add', 'delete', 'add', 'change', 'control']
        authenticated_perms = ["inherit"]
        rdf_type = 'hd:circle'


class Batch(Model):
    invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE, related_name='batches')
    title = models.CharField(max_length=255, blank=True, null=True)

    class Meta(Model.Meta):
        serializer_fields = ['@id', 'title', 'invoice', 'tasks']
        anonymous_perms = ['view', 'add']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']
        depth = 1


class CircleMember(Model):
    circle = models.ForeignKey(Circle, on_delete=models.CASCADE, related_name='members')
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="circles")
    is_admin = models.BooleanField(default=False)

    class Meta(Model.Meta):
        container_path = "circle-members/"
        anonymous_perms = ['view', 'add', 'delete', 'add', 'change', 'control']
        authenticated_perms = ['inherit']
        unique_together = ['user', 'circle']
        rdf_type = 'hd:circlemember'


class Task(models.Model):
    batch = models.ForeignKey(Batch, on_delete=models.CASCADE, related_name='tasks')
    title = models.CharField(max_length=255)

    class Meta(Model.Meta):
        serializer_fields = ['@id', 'title', 'batch']
        anonymous_perms = ['view']
        authenticated_perms = ['inherit', 'add']
        owner_perms = ['inherit', 'change', 'delete', 'control']


class Project(Model):
    description = models.CharField(max_length=255, null=True, blank=False)
    team = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True, related_name='projects')

    class Meta(Model.Meta):
        anonymous_perms = ['view', 'add', 'delete', 'add', 'change', 'control']
        authenticated_perms = ["inherit"]
        rdf_type = 'hd:project'


class DateModel(Model):
    excluded = models.CharField(max_length=255, null=True, default='test')
    value = models.DateField()

    class Meta(Model.Meta):
        rdf_type = "hd:date"
        serializer_fields_exclude = ['excluded']


class DateChild(Model):
    parent = models.ForeignKey(DateModel, on_delete=models.CASCADE, related_name='children')

    class Meta(Model.Meta):
        rdf_type = 'hd:datechild'


@receiver(post_save, sender=User)
def update_perms(sender, instance, created, **kwargs):
    LDPPermissions.invalidate_cache()



from django.db.models import options

__version__ = '1.3.2'
options.DEFAULT_NAMES += (
    'lookup_field', 'rdf_type', 'rdf_context', 'auto_author', 'auto_author_field', 'owner_field', 'view_set',
    'container_path', 'permission_classes', 'serializer_fields', 'serializer_fields_exclude', 'nested_fields',
    'nested_fields_exclude', 'depth', 'anonymous_perms', 'authenticated_perms', 'owner_perms')
default_app_config = 'djangoldp.apps.DjangoldpConfig'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-06-24 17:09
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0012_auto_20200617_1817'),
    ]

    operations = [
        migrations.AddField(
            model_name='follower',
            name='follower',
            field=models.URLField(blank=True, help_text='(optional) the resource/actor following the object'),
        ),
        migrations.AlterField(
            model_name='follower',
            name='inbox',
            field=models.URLField(help_text='the inbox recipient of updates'),
        ),
        migrations.AlterField(
            model_name='follower',
            name='object',
            field=models.URLField(help_text='the object being followed'),
        ),
    ]



import time
from django.conf import settings
from django.contrib.auth.models import _user_get_all_permissions
from django.core.exceptions import PermissionDenied
from django.db.models.base import ModelBase
from rest_framework.permissions import DjangoObjectPermissions
from djangoldp.filters import LDPPermissionsFilterBackend


class LDPPermissions(DjangoObjectPermissions):
    # *DEFAULT* permissions for anon, auth and owner statuses
    anonymous_perms = ['view']
    authenticated_perms = ['inherit']
    owner_perms = ['inherit']
    # filter backends associated with the permissions class. This will be used to filter queryset in the (auto-generated)
    # view for a model, and in the serializing nested fields
    filter_backends = [LDPPermissionsFilterBackend]

    perms_cache = {
        'time': time.time()
    }
    with_cache = getattr(settings, 'PERMISSIONS_CACHE', True)

    @classmethod
    def invalidate_cache(cls):
        cls.perms_cache = {
            'time': time.time()
        }

    @classmethod
    def refresh_cache(cls):
        if (time.time() - cls.perms_cache['time']) > 5:
            cls.invalidate_cache()

    @classmethod
    def is_owner(cls, user, model, obj):
        return obj and hasattr(model._meta, 'owner_field') and (
                getattr(obj, getattr(model._meta, 'owner_field')) == user
                or (hasattr(user, 'urlid') and getattr(obj, getattr(model._meta, 'owner_field')) == user.urlid)
                or getattr(obj, getattr(model._meta, 'owner_field')) == user.id)

    def _get_cache_key(self, model_name, user, obj):
        user_key = 'None' if user is None else user.id
        obj_key = 'None' if obj is None else obj.id
        return 'User{}{}{}'.format(user_key, model_name, obj_key)

    @classmethod
    def get_model_level_perms(cls, model, user, obj=None):
        '''Auxiliary function returns the model-level anon-auth-owner permissions for a given, model, user and object'''
        anonymous_perms = getattr(model._meta, 'anonymous_perms', cls.anonymous_perms)
        authenticated_perms = getattr(model._meta, 'authenticated_perms', cls.authenticated_perms)
        owner_perms = getattr(model._meta, 'owner_perms', cls.owner_perms)

        # 'inherit' permissions means inherit the permissions from the next level 'down'
        if 'inherit' in authenticated_perms:
            authenticated_perms = authenticated_perms + list(set(anonymous_perms) - set(authenticated_perms))
        if 'inherit' in owner_perms:
            owner_perms = owner_perms + list(set(authenticated_perms) - set(owner_perms))

        # apply user permissions and return
        perms = set()
        if user.is_anonymous:
            perms = perms.union(set(anonymous_perms))
        else:
            if cls.is_owner(user, model, obj):
                perms = perms.union(set(owner_perms))
            else:
                perms = perms.union(set(authenticated_perms))
        return perms

    def user_permissions(self, user, obj_or_model, obj=None):
        """
            Filter user permissions for a model class
        """
        self.refresh_cache()
        # this may be a permission for the model class, or an instance
        if isinstance(obj_or_model, ModelBase):
            model = obj_or_model
        else:
            obj = obj_or_model
            model = obj_or_model.__class__
        model_name = model._meta.model_name

        perms_cache_key = self.cache_key(model, obj, user)
        if self.with_cache and perms_cache_key in self.perms_cache:
            return self.perms_cache[perms_cache_key]

        # return permissions - using set to avoid duplicates
        perms = self.get_model_level_perms(model, user, obj)

        if obj is not None and not user.is_anonymous:
            # get permissions from all backends and then remove model name from the permissions
            forbidden_string = "_" + model_name
            perms = perms.union(set([p.replace(forbidden_string, '') for p in _user_get_all_permissions(user, obj)]))

        self.perms_cache[perms_cache_key] = list(perms)

        return self.perms_cache[perms_cache_key]

    def cache_key(self, model, obj, user):
        model_name = model._meta.model_name
        user_key = 'None' if user is None else user.id
        obj_key = 'None' if obj is None else obj.id
        perms_cache_key = 'User{}{}{}'.format(user_key, model_name, obj_key)
        return perms_cache_key

    def filter_user_perms(self, context, obj_or_model, permissions):
        # Only used on Model.get_permissions to translate permissions to LDP
        return [perm for perm in permissions if perm in self.user_permissions(context['request'].user, obj_or_model)]

    # perms_map defines the permissions required for different methods
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': [],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }

    @classmethod
    def get_permissions(cls, method, obj):
        """
            Translate perms_map to request
        """
        kwargs = {
            'app_label': obj._meta.app_label,
            'model_name': obj._meta.model_name
        }

        # Only allows methods that are on perms_map
        if method not in cls.perms_map:
            raise PermissionDenied

        return [perm % kwargs for perm in cls.perms_map[method]]

    def has_permission(self, request, view):
        """
            Access to containers
        """
        from djangoldp.models import Model

        if self.is_a_container(request._request.path):
            try:
                obj = Model.resolve_parent(request.path)
                model = view.parent_model
            except:
                obj = None
                model = view.model
        else:
            obj = Model.resolve_id(request._request.path)
            model = view.model

        # get permissions required
        perms = LDPPermissions.get_permissions(request.method, model)
        user_perms = self.user_permissions(request.user, model, obj)

        # compare them with the permissions I have
        for perm in perms:
            if not perm.split('.')[-1].split('_')[0] in user_perms:
                return False

        return True

    def is_a_container(self, path):
        from djangoldp.models import Model
        container, id = Model.resolve(path)
        return id is None

    def has_object_permission(self, request, view, obj):
        """
            Access to objects
            User have permission on request: Continue
            User does not have permission:   403
        """
        # get permissions required
        perms = LDPPermissions.get_permissions(request.method, obj)
        model = obj
        user_perms = self.user_permissions(request.user, model, obj)

        return LDPPermissions.compare_permissions(perms, user_perms)

    @classmethod
    def has_model_view_permission(cls, request, model):
        '''
        shortcut to compare the requested user's permissions on the model-level
        :return: True or False
        '''
        # compare required permissions with those I have (on the model)
        perms = LDPPermissions.get_permissions('GET', model)
        user_perms = LDPPermissions.get_model_level_perms(model, request.user)
        return cls.compare_permissions(perms, user_perms)

    @classmethod
    def compare_permissions(self, perms, user_perms):
        # compare them with the permissions I have
        for perm in perms:
            if not perm.split('.')[-1].split('_')[0] in user_perms:
                return False
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-08-28 12:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LDPSource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('container', models.URLField()),
                ('federation', models.CharField(max_length=255)),
            ],
            options={
                'ordering': ('federation',),
                'permissions': (('view_source', 'acl:Read'), ('control_source', 'acl:Control')),
            },
        ),
    ]



from django.test import TestCase
from rest_framework.test import APIRequestFactory, APIClient

from djangoldp.tests.models import Post


class TestDelete(TestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()

    def tearDown(self):
        pass

    def test_delete(self):
        post = Post.objects.create(content="content")
        response = self.client.delete('/posts/{}/'.format(post.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 204)

        self.assertEqual(Post.objects.filter(pk=post.pk).count(), 0)



import re
from importlib import import_module

from django.conf import settings


class WebFinger(object):

    def uri(self, name, *args):
        domain = settings.BASE_URL
        return "{domain}{name}".format(domain=domain, name=name)

    def response(self, response_dict, rel, acct):
        return response_dict


for package in settings.DJANGOLDP_PACKAGES:
    try:
        import_module('{}.endpoints.webfinger'.format(package))
    except ModuleNotFoundError:
        pass

model_classes = {cls.__name__: cls for cls in WebFinger.__subclasses__()}

ACCT_RE = re.compile(
    r'(?:acct:)?(?P<userinfo>[\w.!#$%&\'*+-/=?^_`{|}~]+)@(?P<host>[\w.:-]+)')


class Acct(object):
    def __init__(self, acct):
        m = ACCT_RE.match(acct)
        if not m:
            raise ValueError('invalid acct format')
        (userinfo, host) = m.groups()
        self.userinfo = userinfo
        self.host = host


class WebFingerEndpoint(object):
    """
    WebFinger endpoint
    See https://tools.ietf.org/html/rfc7033
    """

    def __init__(self, request):
        self.request = request
        self.params = {}
        self.acct = None

        self._extract_params()

    def _extract_params(self):
        # Because in this endpoint we handle both GET
        # and POST request.
        query_dict = (self.request.POST if self.request.method == 'POST'
                      else self.request.GET)

        self.params['resource'] = query_dict.get('resource', None)
        self.params['rel'] = query_dict.get('rel', None)

    def validate_params(self):
        """
        A resource must be set.
        """

        if self.params['resource'] is None:
            raise WebFingerError('invalid_request')

        try:
            self.acct = Acct(self.params['resource'])
        except ValueError:
            raise WebFingerError('invalid_acct_format')

    def response(self):
        """
        This endpoint only reply to rel="http://openid.net/specs/connect/1.0/issuer"
        :return: a dict representing the Json response
        """

        dict = {
            'subject': self.params['resource'],
            'aliases': [],
            'links': []
        }

        for class_name in model_classes:
            model_class = model_classes[class_name]
            webfinger = model_class()
            dict = webfinger.response(dict, self.params['rel'], self.acct)

        return dict


class WebFingerError(Exception):
    _errors = {
        'invalid_request': "The request provider parameter must contains an url or an email",
        'invalid_acct_format': "Invalid acct format"
    }

    def __init__(self, error=None, dict=None):
        if dict is None:
            self.error = error
            self.description = self._errors.get(error)
        else:
            self.error = dict['error']
            self.description = dict['error_description']

    def create_dict(self):
        dic = {
            'error': self.error,
            'error_description': self.description,
        }

        return dic



import uuid
import time
from django.contrib.auth import get_user_model
from django.test import override_settings
from rest_framework.test import APIClient, APITestCase
from djangoldp.tests.models import Circle, Project
from djangoldp.models import Activity, ScheduledActivity
from djangoldp.activities.services import BACKLINKS_ACTOR, ActivityPubService, ActivityQueueService


class TestsBacklinksService(APITestCase):

    def setUp(self):
        self.client = APIClient(enforce_csrf_checks=True)
        self.local_user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                               password='glass onion')

    def _get_random_external_user(self):
        '''Auxiliary function creates a user with random external urlid and returns it'''
        username = str(uuid.uuid4())
        email = username + '@test.com'
        urlid = 'https://distant.com/users/' + username
        return get_user_model().objects.create_user(username=username, email=email, password='test', urlid=urlid)

    # TODO: inbox discovery (https://git.startinblox.com/djangoldp-packages/djangoldp/issues/233)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_local_object_with_distant_foreign_key(self):
        # a local Circle with a distant owner
        local_circle = Circle.objects.create(description='Test')
        external_user = self._get_random_external_user()
        local_circle.owner = external_user
        local_circle.save()

        # assert that a activity was sent
        self.assertEqual(Activity.objects.all().count(), 1)

        # reset to a local user, another (update) activity should be sent
        local_circle.owner = self.local_user
        local_circle.save()
        self.assertEqual(Activity.objects.all().count(), 2)

        # external user should no longer be following the object. A further update should not send an activity
        # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/257
        '''another_user = get_user_model().objects.create_user(username='test', email='test@test.com',
                                                            password='glass onion')
        local_circle.owner = another_user
        local_circle.save()
        self.assertEqual(Activity.objects.all().count(), 2)'''

        # re-add the external user as owner
        local_circle.owner = external_user
        local_circle.save()

        # delete parent
        local_circle.delete()
        self.assertEqual(Activity.objects.all().count(), 4)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_local_object_with_external_m2m_join_leave(self):
        # a local project with three distant users
        project = Project.objects.create(description='Test')
        external_a = self._get_random_external_user()
        external_b = self._get_random_external_user()
        external_c = self._get_random_external_user()
        project.team.add(external_a)
        project.team.add(external_b)
        project.team.add(external_c)
        self.assertEqual(Activity.objects.all().count(), 3)

        # remove one individual
        project.team.remove(external_a)
        self.assertEqual(Activity.objects.all().count(), 4)

        # clear the rest
        project.team.clear()
        self.assertEqual(Activity.objects.all().count(), 6)
        prior_count = Activity.objects.all().count()

        # once removed I should not be following the object anymore
        project.delete()
        self.assertEqual(Activity.objects.all().count(), prior_count)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_local_object_with_external_m2m_delete_parent(self):
        project = Project.objects.create(description='Test')
        external_a = self._get_random_external_user()
        project.team.add(external_a)
        prior_count = Activity.objects.all().count()

        project.delete()
        self.assertEqual(Activity.objects.all().count(), prior_count + 1)

    # test that older ScheduledActivity is discarded for newer ScheduledActivity
    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_old_invalid_scheduled_activity_discarded(self):

        def send_two_activities_and_assert_old_discarded(obj):
            # there are two scheduled activities with the same object, (and different time stamps)
            old_activity = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='old')
            old_scheduled = ActivityQueueService._save_sent_activity(old_activity, ScheduledActivity)

            new_activity = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Update', summary='new')
            new_scheduled = ActivityQueueService._save_sent_activity(new_activity, ScheduledActivity)

            # both are sent to the ActivityQueueService
            ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', old_scheduled)
            time.sleep(0.1)
            ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', new_scheduled)

            time.sleep(0.1)
            # assert that all scheduled activities were cleaned up
            self.assertEquals(ScheduledActivity.objects.count(), 0)

            # assert that ONLY the newly scheduled activity was sent
            activities = Activity.objects.all()
            self.assertEquals(Activity.objects.count(), 1)
            astream = activities[0].to_activitystream()
            self.assertEquals(astream['summary'], new_activity['summary'])
            activities[0].delete()

        # variation using expanded syntax
        obj = {
            '@id': 'https://test.com/users/test/'
        }
        send_two_activities_and_assert_old_discarded(obj)

        # variation using id-only syntax
        obj = 'https://test.com/users/test/'
        send_two_activities_and_assert_old_discarded(obj)

    # test that older ScheduledActivity is still sent if it's on a different object
    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_old_valid_scheduled_activity_sent(self):
        # there are two scheduled activities with different objects
        obj = 'https://test.com/users/test1/'
        activity_a = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='A')
        scheduled_a = ActivityQueueService._save_sent_activity(activity_a, ScheduledActivity)

        obj = 'https://test.com/users/test2/'
        activity_b = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='B')
        scheduled_b = ActivityQueueService._save_sent_activity(activity_b, ScheduledActivity)

        # both are sent to the same inbox
        ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', scheduled_a)
        ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', scheduled_b)

        # assert that both scheduled activities were sent, and the scheduled activities were cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 2)

    # variation on the previous test where the two activities are working on different models (using the same object)
    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_old_valid_scheduled_activity_sent_same_object(self):
        obj = 'https://test.com/users/test1/'
        target = {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/4/'}
        activity_a = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Add', summary='A', target=target)
        scheduled_a = ActivityQueueService._save_sent_activity(activity_a, ScheduledActivity)

        obj = 'https://test.com/users/test1/'
        target = {'@type': 'hd:joboffer', '@id': 'https://api.test1.startinblox.com/job-offers/1/'}
        activity_b = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Add', summary='B', target=target)
        scheduled_b = ActivityQueueService._save_sent_activity(activity_b, ScheduledActivity)

        # both are sent to the same inbox
        ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', scheduled_a)
        ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', scheduled_b)

        # assert that both scheduled activities were sent, and the scheduled activities were cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 2)

    # variation using an Add and a Remove (one defines target, the other origin)
    # also tests that an unnecessary add is not sent
    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_matching_origin_and_target_not_sent(self):
        a = {'type': 'Add', 'actor': {'type': 'Service', 'name': 'Backlinks Service'},
             'object': {'@type': 'foaf:user', '@id': 'https://api.test2.startinblox.com/users/calum/'},
             'target': {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/3/'}}
        scheduled_a = ActivityQueueService._save_sent_activity(a, ScheduledActivity)
        b = {'type': 'Remove', 'actor': {'type': 'Service', 'name': 'Backlinks Service'},
             'object': {'@type': 'foaf:user', '@id': 'https://api.test2.startinblox.com/users/calum/'},
             'origin': {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/3/'}}
        scheduled_b = ActivityQueueService._save_sent_activity(b, ScheduledActivity)

        # both are sent to the same inbox
        ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', scheduled_a)
        ActivityQueueService._activity_queue_worker('http://127.0.0.1:8001/idontexist/', scheduled_b)

        # assert that both scheduled activities were sent, and the scheduled activities were cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 1)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_unnecessary_add_not_sent(self):
        # an add activity was sent previously
        a = {'type': 'Add', 'actor': {'type': 'Service', 'name': 'Backlinks Service'},
             'object': {'@type': 'foaf:user', '@id': 'https://api.test2.startinblox.com/users/calum/'},
             'target': {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/3/'}}
        ActivityQueueService._save_sent_activity(a, Activity, success=True, type='add',
                                                 external_id='https://distant.com/inbox/')

        # no remove has since been sent, but a new Add is scheduled
        scheduled_b = ActivityQueueService._save_sent_activity(a, ScheduledActivity, success=False, type='add',
                                                               external_id='https://distant.com/inbox/')
        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled_b)

        # assert that only the previous activity was sent, and the scheduled activites cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 1)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_unnecessary_remove_not_sent(self):
        # an remove activity was sent previously
        a = {'type': 'Remove', 'actor': {'type': 'Service', 'name': 'Backlinks Service'},
             'object': {'@type': 'foaf:user', '@id': 'https://api.test2.startinblox.com/users/calum/'},
             'target': {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/3/'}}
        ActivityQueueService._save_sent_activity(a, Activity, success=True, type='remove',
                                                 external_id='https://distant.com/inbox/')

        # no add has since been sent, but a new Remove is scheduled
        scheduled_b = ActivityQueueService._save_sent_activity(a, ScheduledActivity, success=False, type='remove',
                                                               external_id='https://distant.com/inbox/')
        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled_b)

        # assert that only the previous activity was sent, and the scheduled activites cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 1)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_necessary_add_sent(self):
        # a remove activity was sent previously
        a = {'type': 'Remove', 'actor': {'type': 'Service', 'name': 'Backlinks Service'},
             'object': {'@type': 'foaf:user', '@id': 'https://api.test2.startinblox.com/users/calum/'},
             'target': {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/3/'}}
        ActivityQueueService._save_sent_activity(a, Activity, success=True, type='remove',
                                                 external_id='https://distant.com/inbox/')

        # an add is now being sent
        scheduled_b = ActivityQueueService._save_sent_activity(a, ScheduledActivity, type='add',
                                                               external_id='https://distant.com/inbox/')
        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled_b)

        # assert that both activities sent, and the scheduled activites cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 2)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_first_add_sent(self):
        # no activity has been sent with this target, before this add
        a = {'type': 'Add', 'actor': {'type': 'Service', 'name': 'Backlinks Service'},
             'object': {'@type': 'foaf:user', '@id': 'https://api.test2.startinblox.com/users/calum/'},
             'target': {'@type': 'hd:skill', '@id': 'https://api.test1.startinblox.com/skills/3/'}}
        scheduled = ActivityQueueService._save_sent_activity(a, ScheduledActivity, success=True, type='add',
                                                             external_id='https://distant.com/inbox/')
        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled)

        # assert that the activity was sent, and the scheduled activites cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 1)

    # validate Update activity objects have new info before sending the notification
    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_unnecessary_update_not_sent(self):
        # an object was sent in one activity
        obj = {
            '@type': 'hd:circle',
            '@id': 'https://test.com/circles/8/',
            'owner': {'@id': 'https://distant.com/users/john/',
                      '@type': 'foaf:user'}
        }
        activity_a = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='A')
        ActivityQueueService._save_sent_activity(activity_a, Activity, success=True, type='create',
                                                 external_id='https://distant.com/inbox/')

        # now I'm sending an update, which doesn't change anything about the object
        activity_b = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='B')
        scheduled_b = ActivityQueueService._save_sent_activity(activity_b, ScheduledActivity, type='update',
                                                               external_id='https://distant.com/inbox/')

        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled_b)

        # assert that only the previous activity was sent, and the scheduled activites cleaned up
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 1)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_necessary_update_is_sent(self):
        # an object was sent in one activity
        obj = {
            '@type': 'hd:circle',
            '@id': 'https://test.com/circles/8/',
            'owner': {'@id': 'https://distant.com/users/john/',
                      '@type': 'foaf:user'}
        }
        activity_a = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='A')
        ActivityQueueService._save_sent_activity(activity_a, Activity, success=True, type='create',
                                                 external_id='https://distant.com/inbox/')

        # now I'm sending an update, which changes the owner of the circle
        obj['owner']['@id'] = 'https://distant.com/users/mark/'
        activity_b = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='B')
        scheduled_b = ActivityQueueService._save_sent_activity(activity_b, ScheduledActivity, type='update',
                                                               external_id='https://distant.com/inbox/')

        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled_b)

        # assert that both activities were sent
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 2)

    @override_settings(SEND_BACKLINKS=True, DISABLE_OUTBOX='DEBUG')
    def test_first_update_is_sent(self):
        # no prior activity was sent for this object - should send
        obj = {
            '@type': 'hd:circle',
            '@id': 'https://test.com/circles/8/',
            'owner': {'@id': 'https://distant.com/users/john/',
                      '@type': 'foaf:user'}
        }
        activity = ActivityPubService.build_activity(BACKLINKS_ACTOR, obj, activity_type='Create', summary='A')
        scheduled = ActivityQueueService._save_sent_activity(activity, ScheduledActivity, type='update',
                                                             external_id='https://distant.com/inbox/')
        ActivityQueueService._activity_queue_worker('https://distant.com/inbox/', scheduled)
        self.assertEquals(ScheduledActivity.objects.count(), 0)
        self.assertEquals(Activity.objects.count(), 1)



import json
import uuid
from django.contrib.auth import get_user_model
from djangoldp.serializers import LDListMixin, LDPSerializer
from rest_framework.test import APIClient, APITestCase
from guardian.shortcuts import assign_perm

from .models import PermissionlessDummy, Dummy, LDPDummy
from djangoldp.permissions import LDPPermissions


class TestsGuardian(APITestCase):

    def setUp(self):
        self.client = APIClient(enforce_csrf_checks=True)
        LDPPermissions.invalidate_cache()
        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

    def setUpLoggedInUser(self):
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.client.force_authenticate(user=self.user)
        LDPPermissions.invalidate_cache()
        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

    def _get_dummy_with_perms(self, perms=None, parent=None):
        if perms is None:
            perms = []
        dummy = PermissionlessDummy.objects.create(some='test', slug=uuid.uuid4(), parent=parent)
        model_name = PermissionlessDummy._meta.model_name

        for perm in perms:
            assign_perm(perm + '_' + model_name, self.user, dummy)

        return dummy

    # optional setup for testing PermissionlessDummy model with parameterised perms
    def setUpGuardianDummyWithPerms(self, perms=None, parent=None):
        self.dummy = self._get_dummy_with_perms(perms, parent)

    # test that dummy with no permissions set returns no results
    def test_get_dummy_no_permissions(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms()
        response = self.client.get('/permissionless-dummys/{}/'.format(self.dummy.slug))
        self.assertEqual(response.status_code, 403)

    # test with anonymous user
    def test_get_dummy_anonymous_user(self):
        self.setUpGuardianDummyWithPerms()
        response = self.client.get('/permissionless-dummys/')
        self.assertEqual(response.status_code, 403)

    def test_get_dummy_permission_granted(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view'])
        response = self.client.get('/permissionless-dummys/{}/'.format(self.dummy.slug))
        self.assertEqual(response.status_code, 200)

    def test_get_dummy_permission_rejected(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view'])
        dummy_without = PermissionlessDummy.objects.create(some='test2', slug='test2')
        response = self.client.get('/permissionless-dummys/{}/'.format(dummy_without.slug))
        self.assertEqual(response.status_code, 403)

    def test_patch_dummy_permission_granted(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['view', 'change'])
        body = {'some': "some_new"}
        response = self.client.patch('/permissionless-dummys/{}/'.format(self.dummy.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

    def test_patch_dummy_permission_rejected(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['change'])
        dummy_without = PermissionlessDummy.objects.create(some='test2', slug='test2')
        body = {'some': "some_new"}
        response = self.client.patch('/permissionless-dummys/{}/'.format(dummy_without.slug), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 403)

    # test that custom permissions are returned on a model
    def test_custom_permissions(self):
        self.setUpLoggedInUser()
        self.setUpGuardianDummyWithPerms(['custom_permission'])

        permissions = LDPPermissions()
        result = permissions.user_permissions(self.user, self.dummy)
        self.assertIn('custom_permission', result)

    # test that duplicate permissions aren't returned
    def test_no_duplicate_permissions(self):
        self.setUpLoggedInUser()
        dummy = Dummy.objects.create(some='test', slug='test')
        model_name = Dummy._meta.model_name

        assign_perm('view_' + model_name, self.user, dummy)

        permissions = LDPPermissions()
        result = permissions.user_permissions(self.user, dummy)
        self.assertEqual(result.count('view'), 1)

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/297
    '''def test_list_dummy_exception(self):
        self.setUpLoggedInUser()
        # I have permission on a permissionless dummy, but not in general
        dummy_a = self._get_dummy_with_perms()
        dummy_b = self._get_dummy_with_perms(['view'])
        response = self.client.get('/permissionless-dummys/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        self.assertNotIn(response.data['ldp:contains'], dummy_a.urlid)
        self.assertIn(response.data['ldp:contains'], dummy_b.urlid)'''

    def test_list_dummy_exception_nested_view(self):
        self.setUpLoggedInUser()
        parent = LDPDummy.objects.create(some="test")
        # two dummies, one I have permission to view and one I don't
        dummy_a = self._get_dummy_with_perms(parent=parent)
        dummy_b = self._get_dummy_with_perms(['view'], parent)
        response = self.client.get('/ldpdummys/{}/anons/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

    def test_list_dummy_exception_nested_serializer(self):
        self.setUpLoggedInUser()
        parent = LDPDummy.objects.create(some="test")
        # two dummies, one I have permission to view and one I don't
        dummy_a = self._get_dummy_with_perms(parent=parent)
        dummy_b = self._get_dummy_with_perms(['view'], parent)
        response = self.client.get('/ldpdummys/{}/'.format(parent.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['anons']['ldp:contains']), 1)




from django.contrib.auth import get_user_model
from rest_framework.test import APIClient, APITestCase
from djangoldp.tests.models import User, Project
import cProfile, pstats, io


class TestPerformance(APITestCase):
    fixtures = ['test.json',]

    def setUp(self):
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.client.force_authenticate(user=self.user)
        print('there are ' + str(Project.objects.count()) + ' projects in the database')
        print('there are ' + str(User.objects.count()) + ' users in the database')

    def _print_stats(self, pr):
        s = io.StringIO()
        ps = pstats.Stats(pr, stream=s)
        ps.print_stats()
        print(s.getvalue())

    def _enable_new_profiler(self):
        pr = cProfile.Profile()
        pr.enable()
        return pr

    def test_get_container(self):
        # pr = self._enable_new_profiler()
        response = self.client.get('/projects/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        print('counted ' + str(len(response.data['ldp:contains'])) + ' projects')
        # pr.disable()
        #self._print_stats(pr)

        # pr = self._enable_new_profiler()
        response = self.client.get('/users/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        print('counted ' + str(len(response.data['ldp:contains'])) + ' users')
        # pr.disable()
        #self._print_stats(pr)



from django.test import TestCase

from djangoldp.models import Model
from djangoldp.tests.models import Dummy, LDPDummy, Circle, CircleMember


class LDPModelTest(TestCase):

    def test_class_not_inheriting_ldp_model(self):
        dummy = Dummy.objects.create(some="text")
        self.assertEquals("/dummys/", Model.container_id(dummy))
        self.assertEquals("/dummys/{}/".format(dummy.slug), Model.resource_id(dummy))

    def test_class_inheriting_ldp_model(self):
        dummy = LDPDummy.objects.create(some="text")
        self.assertEquals("/ldpdummys/", dummy.get_container_id())
        self.assertEquals("http://happy-dev.fr/ldpdummys/{}/".format(dummy.pk), dummy.get_absolute_url())
        self.assertEquals("/ldpdummys/", Model.container_id(dummy))
        self.assertEquals("/ldpdummys/{}/".format(dummy.pk), Model.resource_id(dummy))

    def test_from_resolve_id(self):
        saved_instance = Dummy.objects.create(some="text", slug="someid")
        result = Model.resolve_id("/dummys/{}/".format(saved_instance.slug))
        self.assertEquals(saved_instance, result)

    def test_resolve_container(self):
        result = Model.resolve_container("/dummys/")
        self.assertEquals(Dummy, result)

    def test_auto_url(self):
        from django.urls import get_resolver
        dummy = LDPDummy.objects.create(some="text")
        view_name = '{}-list'.format(dummy._meta.object_name.lower())
        path = 'http://happy-dev.fr/{}{}/'.format(get_resolver().reverse_dict[view_name][0][0][0], dummy.pk)

        self.assertEquals(path, dummy.get_absolute_url())

    def test_ldp_manager_local_objects(self):
        local = LDPDummy.objects.create(some='text')
        external = LDPDummy.objects.create(some='text', urlid='https://distant.com/ldpdummys/1/')
        self.assertEqual(LDPDummy.objects.count(), 2)
        local_queryset = LDPDummy.objects.local()
        self.assertEqual(local_queryset.count(), 1)
        self.assertIn(local, local_queryset)
        self.assertNotIn(external, local_queryset)

    def test_ldp_manager_nested_fields_auto(self):
        nested_fields = Circle.objects.nested_fields()
        expected_nested_fields = ['team', 'members']
        self.assertEqual(len(nested_fields), len(expected_nested_fields))
        for expected in expected_nested_fields:
            self.assertIn(expected, nested_fields)

        nested_fields = CircleMember.objects.nested_fields()
        expected_nested_fields = []
        self.assertEqual(nested_fields, expected_nested_fields)

    def test_ldp_manager_nested_fields_exclude(self):
        setattr(Circle.Meta, 'nested_fields_exclude', ['team'])
        nested_fields = Circle.objects.nested_fields()
        expected_nested_fields = ['members']
        self.assertEqual(nested_fields, expected_nested_fields)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2020-05-05 17:33
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0008_auto_20200501_1207'),
    ]

    operations = [
        migrations.RenameField(
            model_name='activity',
            old_name='backlink_created',
            new_name='is_backlink',
        ),
        migrations.RenameField(
            model_name='ldpsource',
            old_name='backlink_created',
            new_name='is_backlink',
        ),
    ]



from rest_framework.test import APIRequestFactory, APIClient, APITestCase

from djangoldp.models import LDPSource


class TestSource(APITestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()

    def tearDown(self):
        pass

    def test_get_resource(self):
        source = LDPSource.objects.create(federation="source_name", urlid="http://bar.foo/")
        response = self.client.get('/sources/{}/'.format(source.federation), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['@id'], 'http://happy-dev.fr/sources/source_name/')
        self.assertEqual(len(response.data['ldp:contains']), 1)

    def test_get_empty_resource(self):
        response = self.client.get('/sources/{}/'.format('unknown'), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['@id'], 'http://happy-dev.fr/sources/unknown/')
        self.assertEqual(len(response.data['ldp:contains']), 0)



from rest_framework.utils import model_meta


def get_prefetch_fields(model, serializer, depth, prepend_string=''):
    '''
    This method should then be used with queryset.prefetch_related, to auto-fetch joined resources (to speed up nested serialization)
    This can speed up ModelViewSet and LDPViewSet alike by as high a factor as 2
    :param model: the model to be analysed
    :param serializer: an LDPSerializer instance. Used to extract the fields for each nested model
    :param depth: the depth at which to stop the recursion (should be set to the configured depth of the ViewSet)
    :param prepend_string: should be set to the default. Used in recursive calls
    :return: set of strings to prefetch for a given model. Including serialized nested fields and foreign keys recursively
    called on many-to-many fields until configured depth reached
    '''
    # the objective is to build a list of fields and nested fields which should be prefetched for the optimisation
    # of database queries
    fields = set()

    # get a list of all fields which would be serialized on this model
    # TODO: dynamically generating serializer fields is necessary to retrieve many-to-many fields at depth > 0,
    #  but the _all_ default has issues detecting reverse many-to-many fields
    # meta_args = {'model': model, 'depth': 0, 'fields': Model.get_meta(model, 'serializer_fields', '__all__')}
    # meta_class = type('Meta', (), meta_args)
    # serializer = (type(LDPSerializer)('TestSerializer', (LDPSerializer,), {'Meta': meta_class}))()
    serializer_fields = set([f for f in serializer.get_fields()])

    # we are only interested in foreign keys (and many-to-many relationships)
    model_relations = model_meta.get_field_info(model).relations
    for field_name, relation_info in model_relations.items():
        # foreign keys can be added without fuss
        if not relation_info.to_many:
            fields.add((prepend_string + field_name))
            continue

        # nested fields should be added if serialized
        if field_name in serializer_fields:
            fields.add((prepend_string + field_name))

            # and they should also have their immediate foreign keys prefetched if depth not reached
            if depth >= 0:
                new_prepend_str = prepend_string + field_name + '__'
                fields = fields.union(get_prefetch_fields(relation_info.related_model, serializer, depth - 1, new_prepend_str))

    return fields



from djangoldp.serializers import LDListMixin, LDPSerializer
from django.contrib.auth import get_user_model
from datetime import datetime
from rest_framework.test import APIRequestFactory, APIClient, APITestCase

from djangoldp.tests.models import Post, Invoice, JobOffer, Skill, Batch, DateModel, UserProfile


class TestGET(APITestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

    def tearDown(self):
        pass

    def test_get_resource(self):
        post = Post.objects.create(content="content")
        response = self.client.get('/posts/{}/'.format(post.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['content'], "content")
        self.assertIn('author', response.data)

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/293
    '''def test_get_resource_urlid(self):
        user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                    password='glass onion')
        UserProfile.objects.create(user=user)
        post = Post.objects.create(content="content", author=user.userprofile)
        response = self.client.get('/posts/{}/'.format(post.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['content'], "content")
        self.assertEqual(response.data['author'], user.userprofile.urlid)'''

    def test_get_container(self):
        Post.objects.create(content="content")
        # federated object - should not be returned in the container view
        Post.objects.create(content="federated", urlid="https://external.com/posts/1/")
        response = self.client.get('/posts/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('permissions', response.data)
        self.assertEquals(1, len(response.data['ldp:contains']))
        self.assertEquals(2, len(response.data['permissions']))  # read and add

        Invoice.objects.create(title="content")
        response = self.client.get('/invoices/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('permissions', response.data)
        self.assertEquals(1, len(response.data['permissions']))  # read only

    def test_get_empty_container(self):
        Post.objects.all().delete()
        response = self.client.get('/posts/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(0, len(response.data['ldp:contains']))

    def test_get_filtered_fields(self):
        skill = Skill.objects.create(title="Java", obligatoire="ok", slug="1")
        skill2 = Skill.objects.create(title="Java", obligatoire="ok", slug="2")
        skill3 = Skill.objects.create(urlid="http://external/skills/1")
        job = JobOffer.objects.create(title="job", slug="1")
        job.skills.add(skill)
        job.skills.add(skill2)
        job.skills.add(skill3)
        job.save()
        response = self.client.get('/job-offers/{}/'.format(job.slug), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('recent_skills', response.data)
        self.assertEqual(response.data['recent_skills']['@id'], "http://happy-dev.fr/job-offers/1/recent_skills/")
        self.assertEqual(response.data['skills']['ldp:contains'][2]['@id'], "http://external/skills/1")

    def test_get_reverse_filtered_fields(self):
        skill = Skill.objects.create(title="Java", obligatoire="ok", slug="1")
        job = JobOffer.objects.create(title="job", slug="1")
        job.skills.add(skill)
        job.save()
        response = self.client.get('/skills/{}/'.format(skill.slug), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('recent_jobs', response.data)
        self.assertEqual(response.data['recent_jobs']['@id'], "http://happy-dev.fr/skills/1/recent_jobs/")

    def test_get_virtual_field(self):
        skill = Skill.objects.create(title="Java", obligatoire="ok", slug="1")
        skill2 = Skill.objects.create(title="Java", obligatoire="ok", slug="2")
        job = JobOffer.objects.create(title="job", slug="1")
        job.skills.add(skill)
        job.skills.add(skill2)
        job.save()
        response = self.client.get('/job-offers/{}/'.format(job.slug), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('some_skill', response.data)
        self.assertEqual(response.data['some_skill']['@id'], skill.urlid)

    def test_get_nested(self):
        invoice = Invoice.objects.create(title="invoice")
        batch = Batch.objects.create(invoice=invoice, title="batch")
        distant_batch = Batch.objects.create(invoice=invoice, title="distant", urlid="https://external.com/batch/1/")
        response = self.client.get('/invoices/{}/batches/'.format(invoice.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEquals(response.data['@id'], 'http://happy-dev.fr/invoices/{}/batches/'.format(invoice.pk))
        self.assertEquals(len(response.data['ldp:contains']), 2)
        self.assertEquals(response.data['ldp:contains'][0]['invoice']['@id'], 'http://happy-dev.fr/invoices/{}/'.format(invoice.pk))

    def test_serializer_excludes(self):
        date = DateModel.objects.create(excluded='test', value=datetime.now())
        response = self.client.get('/dates/{}/'.format(date.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertNotIn('excluded', response.data.keys())

    def test_serializer_excludes_serializer_fields_set_also(self):
        setattr(DateModel._meta, 'serializer_fields', ['value', 'excluded'])
        date = DateModel.objects.create(excluded='test', value=datetime.now())
        response = self.client.get('/dates/{}/'.format(date.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertNotIn('excluded', response.data.keys())



import datetime
import platform

import time
from django.contrib.auth import get_user_model
from rest_framework.test import APIRequestFactory, APIClient, APITestCase
from statistics import mean, variance

from djangoldp.permissions import LDPPermissions
from djangoldp.tests.models import Post, Invoice, JobOffer, Skill, Batch, DateModel, Project, User


class TestPerformanceGET(APITestCase):
    posts = []
    skills = []
    jobs = []
    test_volume = 100
    result_line = []
    withAuth = True
    withPermsCache = True
    # fixtures = ['test.json',]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print("Init", end='', flush=True)

        LDPPermissions.with_cache = cls.withPermsCache

        step = cls.test_volume / 10
        cls.factory = APIRequestFactory()

        for i in range(cls.test_volume):
            get_user_model().objects.create_user(username='john{}'.format(i), email='jlennon{}@beatles.com'.format(i),
                                                         password='glass onion')
            cls.posts.append(Post.objects.create(content="content"))
            cls.skills.append(Skill.objects.create(title="Java", obligatoire="ok", slug=str(i)))

        for i in range(cls.test_volume):
            job = JobOffer.objects.create(title="job", slug=str(i))
            for skill in cls.skills:
                job.skills.add(skill)
            if i % step == 0:
                print(".", end='', flush=True)
            job.save()
            cls.jobs.append(job)

        cls.result_line.append(platform.node())
        cls.result_line.append(datetime.datetime.today().strftime("%b %d %Y %H:%M:%S"))
        cls.result_line.append(cls.withAuth)
        cls.result_line.append(cls.withPermsCache)
        cls.result_line.append(cls.test_volume)
        cls.result_line.append("N/A")
        cls.result_line.append("N/A")
        cls.result_line.append("N/A")
        cls.result_line.append("N/A")
        cls.result_line.append("N/A")
        cls.result_line.append("N/A")
        cls.result_line.append("N/A")

    def setUp(self):
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.client.force_authenticate(user=self.user)
        print('there are ' + str(Project.objects.count()) + ' projects in the database')
        print('there are ' + str(User.objects.count()) + ' users in the database')

    @classmethod
    def tearDownClass(cls):
        import csv
        with open('perf_result.csv', 'a', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',',
                                    quotechar='|', quoting=csv.QUOTE_MINIMAL)
            writer.writerow(cls.result_line)

    def test_get_resource(self):
        times = []

        for post in self.posts:
            start_time = time.time()
            response = self.client.get('/posts/{}/'.format(post.pk), content_type='application/ld+json')
            end_time = time.time()
            times.append(end_time - start_time)

        self.result_line[5] = str(mean(times))
        print("Variance execution time :" + str(variance(times)))

    def test_get_container(self):
        times = []

        for post in self.posts:
            start_time = time.time()
            response = self.client.get('/posts/', content_type='application/ld+json')
            end_time = time.time()
            times.append(end_time - start_time)

        self.result_line[6] = str(mean(times))
        print("Variance execution time :" + str(variance(times)))

    def test_get_filtered_fields(self):
        times = []

        for job in self.jobs:
            start_time = time.time()
            response = self.client.get('/job-offers/{}/'.format(job.slug), content_type='application/ld+json')
            end_time = time.time()
            times.append(end_time - start_time)

        self.result_line[7] = str(mean(times))

        print("Variance execution time :" + str(variance(times)))

    def test_get_reverse_filtered_fields(self):
        times = []

        for skill in self.skills:
            start_time = time.time()
            response = self.client.get('/skills/{}/'.format(skill.slug), content_type='application/ld+json')
            end_time = time.time()
            times.append(end_time - start_time)

        self.result_line[8] = str(mean(times))
        print("Variance execution time :" + str(variance(times)))

    def test_get_nested(self):
        times = []

        for job in self.jobs:
            start_time = time.time()
            response = self.client.get('/jobs/{}/skills'.format(job.slug), content_type='application/ld+json')
            end_time = time.time()
            times.append(end_time - start_time)

        self.result_line[9] = str(mean(times))
        print("Variance execution time :" + str(variance(times)))

    def test_get_users_container(self):
        # pr = self._enable_new_profiler()
        # response = self.client.get('/projects/', content_type='application/ld+json')
        # self.assertEqual(response.status_code, 200)
        # print('counted ' + str(len(response.data['ldp:contains'])) + ' projects')
        # pr.disable()
        #self._print_stats(pr)

        # pr = self._enable_new_profiler()
        start_time = time.time()
        response = self.client.get('/users/', content_type='application/ld+json')
        end_time = time.time()
        self.assertEqual(response.status_code, 200)
        print('counted ' + str(len(response.data['ldp:contains'])) + ' users')
        self.result_line[10] = str(end_time-start_time)
        # pr.disable()
        #self._print_stats(pr)




import json
import argparse
from pathlib import Path
from utils import generate_users, generate_projects

'''
A script which generates and outputs random test data, into a file used by the performance unit tests
for help run python test_data_generator.py -h
'''

parser = argparse.ArgumentParser(description='generates and outputs random test data, into a file used by the performance unit tests')
parser.add_argument(dest='count', metavar='N', type=int, help='the number of users (and projects) to generate')

args = parser.parse_args()
count = args.count

user_template = {
    'model': 'tests.user',
    'pk': 0,
    'fields': {
        'username': 'john',
        'email': 'jlennon@c.coop',
        'password':'glass onion'
    }
}

project_template = {
    'model': 'tests.project',
    'pk': 0,
    'fields': {
        'description': 'Test'
    }
}

fixture = generate_users(count, user_template)
fixtue = generate_projects(count, project_template, fixture=fixture)

with open(Path(__file__).parent / "../fixtures/test.json", 'w') as output:
    json.dump(fixture, output)

print(str(count))
